!*******************************************************
!*
!* MODULE for the algebra routines
!*
!*******************************************************

MODULE algebra

  !* use of the constants modules
  !* (mainly for the idp parameter)
  USE constants

contains

  !*****************************************************
  !*
  !* BiCGSTab algorithm
  !*
  !* Computes the solution of the linear system :
  !*
  !*               A X = B
  !* 
  !* INPUT :
  !* 
  !* a :: matrix
  !* 
  !* b :: Right hand side
  !*
  !* OUTPUT :
  !*
  !* x :: solution
  !*
  !*****************************************************  
  
  subroutine bicgstab(a,x,b)

    IMPLICIT NONE

    !* Variables declaration

    REAL(idp), DIMENSION(:,:) :: a
    REAL(idp), DIMENSION(:) :: x,b
    REAL(idp), DIMENSION(:), ALLOCATABLE :: pj,tmppj, rj, sj, r0star, apj, asj,temp4all
    REAL(idp) :: alpha, beta, omega, residu, residu0, rjr0star,eps

    INTEGER :: i,n,itmax

    !* Code

    !** Variables initialization
    !** and memory allocation

    n = size(x,1)
    itmax=n*10
    eps= 1e-10

    ALLOCATE(pj(n),tmppj(n),rj(n),sj(n),r0star(n),apj(n),asj(n),temp4all(n))

    x = 0.d0
    pj = 0.d0
    tmppj = 0.d0
    rj = 0.d0
    sj = 0.d0
    r0star = 0.d0
    apj = 0.d0
    asj = 0.d0
    temp4all = 0.d0

    !** Beginning of the algorithm

    call MATVEC(a,x,temp4all)

    pj = b - temp4all

    rj = pj

    r0star = 1.d0*n*n

    residu = DOTPRODUCT(rj,rj)

    call MATVEC(a,pj,apj)

    !** Beginning of the loops

    DO i=1, itmax

       rjr0star = DOTPRODUCT(rj,r0star)

       alpha = rjr0star/DOTPRODUCT(apj,r0star)

       sj = rj

       sj = sj - alpha*apj

       call MATVEC(a,sj,asj)

       omega = DOTPRODUCT(asj,sj)/DOTPRODUCT(asj,asj)

       x = x + alpha*pj

       x = x + omega*sj

       rj = sj

       rj = rj - omega*asj

       beta = DOTPRODUCT(rj , r0star)/(rjr0star)*(alpha/omega)

       tmppj = pj

       pj = rj

       pj = pj + beta*tmppj

       pj = pj - beta*omega*apj

       call MATVEC(a,x,temp4all)

       residu = sqrt(DOTPRODUCT(rj,rj))

       call MATVEC( a,pj,apj)

       write(*,*) 'it = ', i, 'residu : ', residu

       if(residu .lt. eps) then
          !** The precision has been reached
          !** Exiting the routine
          exit
       end if

    end DO
    !** End of the loop

    !** Memory deallocation
    deallocate(pj, tmppj, rj, sj, r0star, apj, asj, temp4all)

  END SUBROUTINE bicgstab

  !*****************************************************  
  !*
  !* Matrix Vector product
  !*
  !* This subroutine realized the matrix vector product
  !* for a square matrix
  !*
  !* INPUT
  !*
  !* a :: matrix
  !*
  !* x :: vector that should be multiplied by the matrix
  !*
  !* OUTPUT
  !*
  !* b :: solution vector
  !*
  !***************************************************** 

  SUBROUTINE MATVEC(a,x,b)

    IMPLICIT NONE

    !* Variables declaration

    REAL(idp), DIMENSION(:,:) :: a
    REAL(idp), DIMENSION(:)   :: x,b
    REAL(idp), DIMENSION(size(a,1)) :: temp1, temp2

    INTEGER                 :: i,j
    REAL(8)                 :: temp

    !* Code

    boucle_i : do i=1,size(a,1)
       temp = 0.d0
       boucle_j : do j=1,size(a,2)
          temp = temp + a(i,j)*x(j)
       end do boucle_j
       b(i) = temp
    end do boucle_i

  end SUBROUTINE MATVEC

  !***************************************************** 
  !*
  !* dot product
  !*
  !* This function computes the dot product of two 
  !* vectors.
  !* It is assumed that they have the same size.
  !*
  !* INPUT
  !*
  !* u1 :: First vector
  !*
  !* u2 :: Second vector
  !*
  !* OUTPUT
  !*
  !* DOTPRODUCT :: value of the dot product
  !*
  !*****************************************************

  FUNCTION DOTPRODUCT(u1,u2)

    IMPLICIT NONE

    !* Variables declaration

    REAL(idp), DIMENSION(:) :: u1,u2
    REAL(idp)               :: DOTPRODUCT
    REAL(idp)               :: ddot
    INTEGER               :: i,n

    !* Code

    DOTPRODUCT=0.d0
    boucle_i : do i=1,size(u1,1)
       DOTPRODUCT = DOTPRODUCT + u1(i)*u2(i)
    end do boucle_i
    return

  end FUNCTION DOTPRODUCT

END MODULE algebra
